From 637839f50bc77a550acaa49385e6a7b6a18f8eb0 Mon Sep 17 00:00:00 2001
From: Thomas Munro <thomas.munro@gmail.com>
Date: Thu, 7 Jul 2022 16:28:01 +1200
Subject: [PATCH 2/2] Remove redundant ftruncate() when allocating memory.

In early releases of the DSM infrastructure, it was possible to resize
segments.  That was removed in release 12 by commit 3c60d0fa.  Now the
ftruncate() + posix_fallocate() sequence during DSM segment creation is
has a redundant step: we're always extending from zero to the desired
size, so we might as well just call posix_fallocate().

Let's also include the remaining ftruncate() call (non-Linux POSIX
systems) in the wait event reporting, for good measure.
---
 src/backend/storage/ipc/dsm_impl.c | 48 ++++++++++++++----------------
 1 file changed, 22 insertions(+), 26 deletions(-)

diff --git a/src/backend/storage/ipc/dsm_impl.c b/src/backend/storage/ipc/dsm_impl.c
index aa47e5de70..4a7492b37c 100644
--- a/src/backend/storage/ipc/dsm_impl.c
+++ b/src/backend/storage/ipc/dsm_impl.c
@@ -363,40 +363,36 @@ dsm_impl_posix_resize(int fd, off_t size)
 	 */
 	PG_SETMASK(&BlockSig);
 
-	/* Truncate (or extend) the file to the requested size. */
-	rc = ftruncate(fd, size);
-
+	pgstat_report_wait_start(WAIT_EVENT_DSM_FILL_ZERO_WRITE);
+#if defined(HAVE_POSIX_FALLOCATE) && defined(__linux__)
 	/*
-	 * On Linux, a shm_open fd is backed by a tmpfs file.  After resizing with
-	 * ftruncate, the file may contain a hole.  Accessing memory backed by a
+	 * On Linux, a shm_open fd is backed by a tmpfs file.  If we were to use
+	 * ftruncate, the file would contain a hole.  Accessing memory backed by a
 	 * hole causes tmpfs to allocate pages, which fails with SIGBUS if there
 	 * is no more tmpfs space available.  So we ask tmpfs to allocate pages
 	 * here, so we can fail gracefully with ENOSPC now rather than risking
 	 * SIGBUS later.
+	 *
+	 * We still use a traditional EINTR retry loop to handle SIGCONT.
+	 * posix_fallocate() doesn't restart automatically, and we don't want
+	 * this to fail if you attach a debugger.
 	 */
-#if defined(HAVE_POSIX_FALLOCATE) && defined(__linux__)
-	if (rc == 0)
+	do
 	{
-		/*
-		 * We still use a traditional EINTR retry loop to handle SIGCONT.
-		 * posix_fallocate() doesn't restart automatically, and we don't want
-		 * this to fail if you attach a debugger.
-		 */
-		pgstat_report_wait_start(WAIT_EVENT_DSM_FILL_ZERO_WRITE);
-		do
-		{
-			rc = posix_fallocate(fd, 0, size);
-		} while (rc == EINTR);
-		pgstat_report_wait_end();
+		rc = posix_fallocate(fd, 0, size);
+	} while (rc == EINTR);
 
-		/*
-		 * The caller expects errno to be set, but posix_fallocate() doesn't
-		 * set it.  Instead it returns error numbers directly.  So set errno,
-		 * even though we'll also return rc to indicate success or failure.
-		 */
-		errno = rc;
-	}
-#endif							/* HAVE_POSIX_FALLOCATE && __linux__ */
+	/*
+	 * The caller expects errno to be set, but posix_fallocate() doesn't
+	 * set it.  Instead it returns error numbers directly.  So set errno,
+	 * even though we'll also return rc to indicate success or failure.
+	 */
+	errno = rc;
+#else
+	/* Extend the file to the requested size. */
+	rc = ftruncate(fd, size);
+#endif
+	pgstat_report_wait_end();
 
 	PG_SETMASK(&UnBlockSig);
 
-- 
2.30.2

